# Graph generation

## Installing clingo

We support the use of Clingo. For installation see [link](https://potassco.org/clingo/) or [link](https://github.com/potassco/clingo-cmake-example).
Clingo **must** be installed via conda to ensure that cmake works.


## CaDiCaL version

To use CaDiCaL via the IPASIR-UP interface, it first must be configured and the 
library "libcadical.a" must be built at the following location:

	../cadical-user-propagation/build/libcadical.a

See README.md and BUILD.md files in the folder "cadical-user-propagation" for
more details related to the build process.

## Building graph generation tool

If Cadical and Clingo are installed, one can build the graph genereation tool using `cmake`.  
Use `cmake . -B./build && cd ./build && make` or execute the `buildScript.sh`.

The program is available at `./build/src/graphgen` and for example `./build/src/graphgen -v 7 --allModels --frequency 5 --useCadical --printStats` can be used to generate all graphs up to isomorphism.

The most important arguments of graphgen are as follows:

`./build/graphgen -v VERTICES [--allModels] [--frequency FREQUENCY] [--cnf FILE] [--useCadical] [--assignmentCutoff ACUTOFF] [--assignmentCutoffPrerunTime TIME] [--cutoff CUTOFF] [--printStats]`

- `-v VERTICES` gives the number of vertices (order) of the graph
- `--allModels` if all graphs should be computed
- `--frequency FREQUENCY` used for balancing the time in the minimality check and the solver itself. For example if it the frequency is 5 then the minimality check only called every 5-th time.  
- `--cnf FILE` the file name  providing the SAT encoding assuming that the edge variables are given row wise by the upper triangle. For example if $n=4$ then the corresponding edge variables are given by the following literals: $e_{0,1} \leftrightarrow 1, e_{0,2} \leftrightarrow 2, e_{0,3} \leftrightarrow 3, e_{1,2} \leftrightarrow 4, e_{1,3} \leftrightarrow 5, e_{2,3} \leftrightarrow 6$. This can also represented by the following matrix (ignoring diagonal entries) <br> 
`- 1 2 3 ` <br>
`1 - 4 5 ` <br>
`2 4 - 6 ` <br>
`3 5 6 - ` <br>

- `--useCadical` for using Cadical as solver. By default Clingo is used.
- `--assignmentCutoff ACUTOFF` is used for cubing, if at least `ACUTOFF` edge variables are assigned and propagate is called, then a cube representing the partial assignement is generated.
- `--assignmentCutoffPrerunTime TIME` gives the prerun time before starting cubing
- `--cutoff CUTOFF` gives the maximal recursive calls in the minimality check the avoid some worst cases. Note that this can result in an incomplete symmetry breaking.
- `--printStats` prints some statistics of the computations

For more arguments we recommend looking at `sms-cpp-impl/src/main.cpp`

## Information for reproducing results in IPASIR-UP paper for SMS

We start with describing the arguments which must be added for the 6 different configurations (the can also be added to the script and forwarded to the graphgen program):

Arguments for CaDiCaL:
- `default` uses the following arguments:  ` --useCadical --propagateLiterals --checkSolutionInProp `
- `enum-IPASIR` uses the following arguments: ` --useCadical --propagateLiterals `
- `no-prop` uses the follwoing arguments: ` --useCadical --checkSolutionInProp`
- `no-inpro` uses the follwoing arguments:` --useCadical --propagateLiterals --checkSolutionInProp --turnOffInprocessing`

Arguments for Clingo:
- `red` uses no additional arguments
- `irred` uses the follwoing arguments: ` --irredundantSymClauses`

The script for producing the graphs is in folder `scripts/runScripts`.
For example, for producing all graphs up to isomorphism for 7 vertices execute with CaDiCaL version default then use

`scripts/runScripts/script_all.sh 7 --useCadical --propagateLiterals --checkSolutionInProp`

The first argument is the number of vertices the later are forwarded to graphgen.

Similar for Kochen Specker candidates use the following for producing all KS graph with 16 vertices.

`scripts/runScripts/script_ks.sh 16 --useCadical --propagateLiterals --checkSolutionInProp`








<!---
## Commands used for KS-graphs

The encoding is generated by the script `sms-cpp-impl/scripts/encodings/kochen_specker.py` which can be fed to the graphgeneration tool.
The script `./scripts/runScripts/script_kochen_specker.sh` combines both `graphgen` and the script for generating the encoding.
The syntax is as follows
`script_kochen_specker.sh N TMP_FILE ARGS_PY ARGS_GRAPHGEN  `,
where `N` gives the number of vertices `TMP_FILE` a temporary file for storing the encoding, `ARGS_PY` the arguments for the python script constructing the encoding and `ARGS_GRAPHGEN` the arguments for the graphgeneration tool.


- `./scripts/runScripts/script_kochen_specker.sh 17 tempfile.txt "" "--frequency 30 --allModels --useCadical --printStats --non010"`: produces all canidate solutions for for $n=17$ using Cadical.

- `./scripts/runScripts/script_kochen_specker.sh 17 tempfile.txt "" "--frequency 30 --allModels --printStats --non010"`: produces all canidate solutions for for $n=17$ using Clingo.


- `./scripts/runScripts/script_kochen_specker.sh 22 tempfile.txt "" "--frequency 30 --allModels --printStats --non010 --assignmentCutoffPrerunTime 5 --assignmentCutoff 110 --useCadical" `: produces cubes with at least 110 assigned edge variables and 5 seconds prerun time.

- `./scripts/runScripts/script_kochen_specker.sh 22 tempfile.txt "" "--frequency 30 --allModels --printStats --non010 --useCadical --cubeFile cubeFile --cube2solve 2390 2392" `: solve all cubes given in `cubeFile` between line 2390 and 2392. --->

