#ifndef GLOBAL_H
#define GLOBAL_H

#include "useful.h"

extern int vertices;
extern int b_vertices[2];
extern vector<bool> initialPartition;
extern int lastVertex;
extern int nextFreeVariable;
extern bool allModels;
extern int timeout;

extern bool generate_connected;

extern bool equalityVersion;

extern bool useCadical;

extern vector<vector<lit_t>> edges;
extern vector<vector<lit_t>> edges_intersection_graph;
extern vector<vector<vector<lit_t>>> triangles;
extern FILE *symBreakClausesFile;
extern FILE *clauseFile;
extern int maxPathLength;
extern int frequency;
extern int frequencyConnectedComponentsSwap;
extern bool printIntermediateStatistic;
extern bool printFullMatrix;

extern vector<pair<int, int>> intervallsColoring;
extern int non010colorable;
extern int non010colorableAll;
extern bool hypermode; // for hypergraphs and hyperspace travel
extern int hypercoloring;
extern int hyperedgeColoring;
extern int minChromaticNumber;
extern int minChomaticNumberSubgraph;
extern int fixedSubgraphSize;

extern string cubeFile;
extern pair<int, int> rangeCubes;

#define COUNT_ASSIGNED 0
#define ESTIMATE_SPACE 1
extern int assignmentScoring;
extern int assignmentCutoff;
extern int assignmentCutoffPrerun;
extern long assignmentCutoffPrerunTime;

extern bool addPermutedColorings;
extern bool fiveCycleVersion;

extern int coloringAlgo;

extern pair<int, int> embeddedStar;                              // min and max Degree
extern vector<pair<int, pair<int, int>>> embeddedStarIntervalls; // contain the neighbors of the root of the star, and the intervall of 2nd neighbors
extern int cutoff;                                               // cutoff for minimality check

typedef struct
{
    clock_t start;
    clock_t timePropagator;
    clock_t timeMinimalityCheck;
    clock_t timeMinCheckSwap;
    clock_t timeCheckFullGraphs;
    long long callsPropagator;
    long long callsCheck;
    long long nSymBreakClauses; // number of symmetry breaking clauses
    long long hyperclauses;
} statistics;

extern vector<vector<lit_t>> edge_stats;
extern vector<vector<vector<lit_t>>> triangle_stats;
void recordGraphStats(const adjacency_matrix_t &matrix);

extern statistics stats;
extern bool printStats;
extern bool hideGraphs;

extern const char **optionsClingo;
extern int nOptionsClingo;

extern bool checkComplement;
extern bool turnoffPropagatorClingo;
extern bool turnoffSMS;
extern bool turnOffInprocessing;

extern string proofFile;

extern bool printFullyDefinedGraphs;

extern bool triangleVersion;

extern vector<vector<vertex_t>> initialVertexOrderings;

extern bool checkNotInpropagator; // doen't check solution in propagator but only after solution reconstruction; potetentially allows more variables to be flipped; currently only for Cadical

extern vector<vector<pair<int, int>>> forbiddenSubgraphs; // given as edgelists
extern bool eliminateAllsubgraphs;

extern int printPartiallyDefined;
extern int planarityFrequency;
extern int thickness2Frequency;

extern bool chechSolutionInProp;
extern FILE *addedClauses;
extern bool propagateLiteralsCadical;
extern bool irredundantSymClauses;
#endif
