#ifndef SOLVE_GENERAL_H
#define SOLVE_GENERAL_H

#include "useful.h"

class CommonInterface
{
public:
  void solve(); // TODO suitable return value;
  virtual ~CommonInterface() {}

  // functions which must be implemented for the concrete solver
protected:
  int nModels = 0;

  virtual void solve(vector<int> assumptions) { EXIT_UNWANTED_STATE }              // solve the formula under the assumption
  virtual bool solve(vector<int> assumptions, int timeout) { EXIT_UNWANTED_STATE } // solve with a given timeout; return false if timeout was reached

  virtual void addClause(const vector<lit_t> &clause, bool redundant) { EXIT_UNWANTED_STATE }
  virtual adjacency_matrix_t getAdjacencyMatrix() { EXIT_UNWANTED_STATE }

  adjacency_matrix_t currentMatrix; // current partially defined graph 

  // functions which are the same for all solvers, which use the previous funcitons
private:
  bool checkMin(bool checkConnectedComponents); // checks if adjacency matrix is minimal
  bool cutoffFunction();                        // If certain number of edge variables is assigned, a cube will be generated

  bool checkFullyDefinedGraph(const adjacency_matrix_t &matrix); // check the property of the fully defined graph
  void printStatistics();
  bool checkChromaticNumberSubgraph();
  bool extractForbiddenSubgraphs(adjacency_matrix_t matrix); // searches for forbidden subgraph in the given adjacency matrix, and forbids them in the solver
  bool planarityTesting(); 
  bool thicknessTesting();
public:
  bool propagate(); // Check state of partial assignment and add clauses if necessary; returns true if no clause was added otherwise false
  bool check();     // prints and excludes graph from search space; returns true if no clause was added otherwise false
  bool subgraphWasUpdated = false;
};

#endif
