#ifndef USEFUL_H
#define USEFUL_H

#include <algorithm>
#include <utility>
#include <vector>
#include <iostream>
#include <numeric>
#include <string>
#include <string.h>
#include <cassert>
#include <fstream>
#include <sstream>

using namespace std;

typedef int lit_t;
typedef vector<lit_t> vec;
typedef vector<lit_t> clause_t;
typedef vector<clause_t> cnf_t;

#define PRINT_CURRENT_LINE                            \
    printf("Line %d, file %s\n", __LINE__, __FILE__); \
    fflush(stdout);

#define EXIT_UNWANTED_STATE                                                          \
    {                                                                                \
        printf("Error: unexpected state at line %d, file %s\n", __LINE__, __FILE__); \
        exit(EXIT_FAILURE);                                                          \
    }

typedef int vertex_t;
typedef std::pair<vertex_t, vertex_t> edge_t;

typedef enum
{
    unknown = 2,
    adjacent = 1,
    not_adjacent = 0
} adjacencyState_t;

typedef std::vector<std::vector<adjacencyState_t>> adjacency_matrix_t;
typedef std::vector<std::vector<adjacencyState_t>> incidence_matrix_t;
typedef std::pair<adjacencyState_t, edge_t> signed_edge_t;

void printAdjacencyMatrix(const adjacency_matrix_t &matrix);
void printPartiallyDefinedAdjacencyMatrix(const adjacency_matrix_t &matrix);
void printHypergraph(const adjacency_matrix_t &matrix, int nVertices[2]);


#endif
